import os
import json
import tempfile
from datetime import timedelta, datetime
from pathlib import Path
import time
from typing import Dict, List
import unicodedata
import re

from croniter import croniter
from ruxit.api.base_plugin import RemoteBasePlugin
from ruxit.api.exceptions import NothingToReportException, ConfigException

from dbquery_imports.log_proxy import LogProxy
from dbquery_imports.models import Result, QueryConfig, MintMetric, QueryResult
from dbquery_imports import connector
from dynatrace import Dynatrace


class CustomDBQueryPluginRemote(RemoteBasePlugin):
    def initialize(self, **kwargs):
        self.validate_parameters()

        # Cache to hold queries that were supposed to be sent, but that the result was not yet available
        # This can happen for long-running queries
        self.not_sent_yet: Dict[str, QueryConfig] = {}
        self.already_sent: Dict[str, QueryConfig] = {}
        self.temp_folder = create_or_get_temp_folder()


    def query(self, **kwargs):
        start_time = time.time()

        self.logger.setLevel(self.config.get("log_level"))

        # This determines if we always send metrics every minute, even when queries are not running every minute
        self.send_cached = self.config.get("send_cached", True)

        # These are to tie the metrics to a custom device
        # Useful for management zones and alerting
        group_name = self.config.get("group_name") or "Custom DB Queries Group"
        device_name = self.config.get("custom_device_name") or "Custom DB Queries Device"
        group = self.topology_builder.create_group(group_name, group_name)
        self.custom_device = group.create_device(device_name)
        self.custom_device_id = f"CUSTOM_DEVICE-{self.custom_device.id:X}"

        # Dynatrace API client to report the MINT metrics
        dt = Dynatrace(self.config["dt_api_url"], self.config["dt_api_token"], log=self.logger)

        # This is for us to send statistics about the queries, like response time, status, number of rows, for troubleshooting purposes
        self.send_statistics = self.config.get("send_statistics", False)
        statistics_group = self.topology_builder.create_group("DBQuery Monitoring", "DBQuery Monitoring")
        name = self.activation.entity_id
        if self.activation.endpoint_name:
            name = self.activation.endpoint_name
        self.statistics_device = statistics_group.create_device(
            f"Statistics ({self.activation.entity_id})", f"DB Query Statistics ({name})"
        )

        if self.send_statistics:
            self.statistics_device.report_property("Endpoint name or ID", f"{name}")
            self.statistics_device.report_property("Type", self.config.get("database_type"))
            self.statistics_device.report_property("Host", self.config.get("database_host"))

        # First, figure out which queries to run, based on the schedules
        queries = self.get_queries()
        queries_to_run = [q for q in queries if q.will_run_now]

        if self.send_statistics:
            self.statistics_device.report_property("Queries", f"{len(queries)}")
        if queries_to_run:
            # Write the config only for the queries that must run now, and start the java process that does the querying
            self.write_config(queries_to_run)

            # Extra arguments to pass to the JVM, like -Xmx60m
            jvm_extra_args = []
            if self.config.get("jvm_extra_args"):
                jvm_extra_args = self.config.get("jvm_extra_args").splitlines()

            # Spawn the java process that will run the queries
            # The password is passed as an Environment Variable for security
            env = os.environ.copy()
            env["DATABASE_PASSWORD"] = self.config["database_password"]
            connector.query(
                self.activation.entity_id,
                self.temp_folder,
                env=env,
                java_path=self.config.get("jvm_path"),
                jvm_extra_args=jvm_extra_args,
            )

        # Collect the queries results, send them 1000 MINT lines at a time
        mint_lines = self.parse_results(queries)
        for lines_subset in chunks(mint_lines, 1000):
            try:
                mint_strings = [m.to_mint() for m in lines_subset]
                self.logger.info(f"Reporting lines: {mint_strings}")
                dt.metrics.ingest(mint_strings)
            except Exception as e:
                error_message = f"Could not publish all metrics: {e}"
                self.logger.error(error_message)
                self.statistics_device.report_error_event(error_message, "Custom DB Query Extension error")

        self.parse_logs()

        self.custom_device.absolute("duration", time.time() - start_time)

    def parse_logs(self):
        log = Path(self.temp_folder, "log", f"{self.activation.entity_id}.log")
        cache = Path(self.temp_folder, f"{self.activation.entity_id}.json")

        for line in LogProxy(log, cache).get_lines_from_file():
            self.logger.info(line)
            if "level=error" in line or "level=warn" in line or "level=fatal" in line:
                self.statistics_device.report_custom_info_event(
                    f"Received a bad line from the log: {line}", "Bad log line received", properties={"line": line}
                )

    def get_queries(self) -> List[QueryConfig]:
        """
        Determines which queries need to be run on this interval
        This is based on the schedule of the query, by default a query runs every minute
        :return: A list of parsed QueryConfig objects
        """
        queries = []
        now = datetime.now().astimezone()

        # We have up to 10 queries
        for i in range(1, 11):
            name = self.config.get(f"query_{i}_name", "")

            try:
                if not name:
                    # Queries that don't have a name are ignored
                    continue

                # Default schedule is every minute
                schedule = self.config.get(f"query_{i}_schedule") or "*/1 * * * *"

                query_string = self.config[f"query_{i}_value"]

                # The user can specify key=value dimension pairs
                extra_dimensions = self.config.get(f"query_{i}_extra_dimensions", "")
                if extra_dimensions:
                    extra_dimensions = extra_dimensions.split(",")

                # The user can specify column names to capture as values
                value_columns = []
                value_columns_raw = self.config.get(f"query_{i}_value_columns", "")
                if value_columns_raw:
                    value_columns = value_columns_raw.split(",")

                # The user can specify column names to capture as dimensions
                dimension_columns_raw = self.config.get(f"query_{i}_dimension_columns", "")
                dimension_columns = []
                if dimension_columns_raw:
                    dimension_columns = dimension_columns_raw.split(",")

                # Parses the cron schedule for this query
                cron = croniter(schedule, now, ret_type=datetime)
                previous_execution = cron.get_prev().astimezone()
                next_execution = cron.get_next().astimezone()

                query_config = QueryConfig(
                    name, query_string, value_columns, dimension_columns, extra_dimensions=extra_dimensions, schedule=schedule
                )

                # Figure out if this query needs to run now
                time_since = now - previous_execution
                if time_since <= timedelta(minutes=1):
                    self.logger.info(
                        f"Adding query '{name}' because it has been scheduled to run {time_since} ago ({previous_execution}) "
                    )
                    query_config.will_run_now = True
                    if query_config.name in self.already_sent:
                        del self.already_sent[query_config.name]
                else:
                    self.logger.info(f"Not adding query '{name}'. Next execution at {next_execution}, now is {now}")
                    query_config.will_run_now = False

                queries.append(query_config)
                if self.send_statistics:
                    state = "RUNNING" if query_config.will_run_now else "CACHED"
                    self.statistics_device.state_metric("query_state", state, dimensions={"Query": query_config.name})

            except Exception as e:
                self.logger.error(f"Could not add query {name}, error: {e}")

        return queries

    def build_connection_string(self) -> str:
        if self.config.get("jdbc_connection_string", ""):
            return self.config.get("jdbc_connection_string")

        database_type = self.config["database_type"]
        host = self.config["database_host"]
        port = self.config["database_port"]
        database = self.config["database_name"]

        if database_type == "MySQL":
            return f"jdbc:mysql://{host}:{port}/{database}"

        if database_type == "PostgreSQL":
            return f"jdbc:postgresql://{host}:{port}/{database}"

        if database_type == "Oracle":
            if self.config["oracle_listener_type"] == "SID":
                return f"jdbc:oracle:thin:@{host}:{port}:{database}"
            else:
                return f"jdbc:oracle:thin:@//{host}:{port}/{database}"

        if database_type == "SQL Server":
            return f"jdbc:sqlserver://{host}:{port};databaseName={database}"

        if database_type == "DB2":
            return f"jdbc:db2://{host}:{port}/{database}"

        if database_type == "Informix":
            return f"jdbc:informix-sqli://{host}:{port}/{database}"

        if database_type == "SAP HANA":
            return f"jdbc:sap://{host}:{port}/?encrypt=true&validateCertificate=false"

    def write_config(self, queries_to_run: List[QueryConfig]):
        """
        Writes the database configuration to a json file named after the endpoint ID
        """

        driver_classes = {
            "DB2": "com.ibm.db2.jcc.DB2Driver",
            "SQL Server": "com.microsoft.sqlserver.jdbc.SQLServerDriver",
            "Oracle": "oracle.jdbc.driver.OracleDriver",
            "MySQL": "com.mysql.cj.jdbc.Driver",
            "PostgreSQL": "org.postgresql.Driver",
            "Informix": "com.informix.jdbc.IfxDriver",
            "SAP HANA": "com.sap.db.jdbc.Driver",
        }

        queries_json = [query.__dict__ for query in queries_to_run]
        config = {
            "username": self.config["database_username"],
            "logLevel": self.config["log_level"],
            "queries": queries_json,
            "driverClassName": driver_classes.get(self.config["database_type"]),
            "connectionString": self.build_connection_string(),
        }
        self.logger.info(f"Writing config {config}")

        path = Path(self.temp_folder, "config")
        with open(Path(path, f"{self.activation.entity_id}.json"), "w") as f:
            json.dump(config, f, indent=2)

    def parse_results(self, queries: List[QueryConfig]) -> List[MintMetric]:
        """
        Parses the results of the queries and returns a list of the results.

        The results are stored as a json file in the temp folder.
        There is one file per endpoint ID
        """
        mint_lines = []
        now = datetime.now().astimezone()

        result = self.read_results_from_file()
        self.logger.info(f"Processing result {result}")

        if result.error:
            # This means there was an error connecting to the Database
            # Report the error and the availability
            error_message = f"Error running queries: {result.error_message}"
            self.logger.error(error_message)
            self.custom_device.report_error_event(error_message, "Custom DB Query Extension error")
            if self.send_statistics:
                self.statistics_device.state_metric("result_state", "ERROR")

        else:
            if self.send_statistics:
                self.statistics_device.state_metric("result_state", "RUNNING")
            # If we get here we have successfully queried the database, parse the results
            for query in queries:
                query_result = result.find_query_by_name(query.name)

                # The result file doesn't even exist yet, the query did not have time to run once
                if query_result is None:
                    self.logger.warning(f"Needed to send query {query} but the result is not available yet")
                    self.not_sent_yet[query.name] = query
                    continue

                # The result file exists, but how long ago was the query executed?
                query_timestamp = datetime.fromtimestamp(query_result.timestamp).astimezone()
                time_since_result_arrived = now - query_timestamp

                cron = croniter(query.schedule, now, ret_type=datetime)
                previous_execution = cron.get_prev().astimezone()
                next_execution = cron.get_next().astimezone()
                time_since_last_scheduled_execution = now - previous_execution
                result_older_than_schedule = time_since_result_arrived > time_since_last_scheduled_execution

                # Special case for queries that run every minute
                # For these it can happen that the result was available on the previous minute, and we report it now
                if result_older_than_schedule and \
                        (next_execution - previous_execution) == timedelta(minutes=1) and \
                        time_since_result_arrived <= timedelta(seconds=90):
                    result_older_than_schedule = False

                already_sent = query.name in self.already_sent

                self.logger.debug(
                    f"'{query.name=}', {query_timestamp=:%H:%M:%S}, {time_since_result_arrived.total_seconds()=}, "
                    f"{previous_execution=:%H:%M:%S}, "
                    f"{time_since_last_scheduled_execution.total_seconds()=}, {result_older_than_schedule=}, "
                    f"{already_sent=}"
                )

                # In this case, the query ran more than one minute ago, but the user wants us to send this cached result
                if (
                    self.send_cached
                    and time_since_result_arrived > timedelta(minutes=1)
                    and query.name not in self.not_sent_yet
                ):
                    self.logger.info(f"Sending cached value for '{query.name}'")
                    mint_lines.extend(self.parse_result(query, query_result))
                    continue

                # In this case, this result is too old, but the user does not want to send cached results
                # we need to keep this query in self.not_sent_yet because we are waiting for a more recent result
                elif not self.send_cached and result_older_than_schedule and query.name not in self.already_sent:
                    self.logger.info(
                        f"Adding '{query.name} to not_sent_yet because the results are too old: {time_since_result_arrived} ago"
                    )
                    self.not_sent_yet[query.name] = query
                    continue

                # When we get here, this is either a query that was supposed to run now, and it did
                # Or a query that is in self.not_sent_yet and ran now, we were waiting for this result
                if query.name not in self.already_sent:
                    mint_lines.extend(self.parse_result(query, query_result))
                    if query.name in self.not_sent_yet:
                        del self.not_sent_yet[query.name]
                    self.already_sent[query.name] = query

        return mint_lines

    def parse_result(self, query: QueryConfig, result: QueryResult) -> List[MintMetric]:
        self.logger.info(f"Query: {query}, Result: {result}")
        mint_lines: List[MintMetric] = []
        if result.error:
            title = f"Error during query execution for {query.name}"
            description = f"There was an error obtaining the results. The error was: {result.error_message}"
            self.statistics_device.report_error_event(description, title, properties={"Query": result.name})
            if self.send_statistics:
                self.statistics_device.state_metric("query_state", "ERROR", dimensions={"Query": result.name})

        if self.send_statistics:
            self.statistics_device.absolute("query_duration", result.duration, dimensions={"Query": result.name})
            self.statistics_device.absolute("query_rows", len(result.rows), dimensions={"Query": result.name})

        # These are optional hardcoded dimensions from plugin.json
        extra_dimensions = {}
        for hardcoded_dimension in query.extra_dimensions:
            key, value = hardcoded_dimension.split("=")
            extra_dimensions[key] = value
        extra_dimensions["dt.entity.custom_device"] = self.custom_device_id
        extra_dimensions["query_name"] = query.name

        metric_name = "custom.db.query"
        for row in result.rows:
            try:
                dimensions: Dict[str, str] = extra_dimensions.copy()
                if query.value_columns:
                    # If the user chose which columns to get
                    for c in query.dimension_columns:
                        # If the user chose to get dimension columns
                        col = row.get_column_by_name(c)
                        dimensions[slugify(col.name)] = col.value

                    for column in query.value_columns:
                        # Report dimensions and values
                        col = row.get_column_by_name(column)
                        if dimensions:
                            report_dimensions = dimensions.copy()
                            report_dimensions["column"] = slugify(column)
                            self.logger.debug(
                                f"{query.name} - Reporting dimensions: {report_dimensions}, value: {col.value}"
                            )
                            if not str(col.value).isnumeric():
                                error_message = f"Value '{col.value}' for column '{col.name}' in query '{query.name}' is not numeric, cannot report it"
                                self.custom_device.report_error_event(error_message, "Custom DB Query Extension error - bad value")
                            mint_lines.append(MintMetric(metric_name, col.value, report_dimensions))
                        else:
                            self.logger.debug(
                                f"{query.name} - Reporting value for column {column} (no dimension): {col}"
                            )
                            if not str(col.value).isnumeric():
                                error_message = f"Value '{col.value}' for column '{col.name}' in query '{query.name}' is not numeric, cannot report it"
                                self.custom_device.report_error_event(error_message, "Custom DB Query Extension error - bad value")
                            mint_lines.append(MintMetric(metric_name, col.value, {"column": column}))
                else:
                    # If no columns were selected, report the first column of the first row
                    col = row.columns[0]
                    self.logger.debug(f"{query.name} - No columns set, reporting first column of the first row: {col}")
                    if not str(col.value).isnumeric():
                        error_message = f"Value '{col.value}' for column '{col.name}' in query '{query.name}' is not numeric, cannot report it"
                        self.custom_device.report_error_event(error_message, "Custom DB Query Extension error - bad value")
                    if dimensions:
                        mint_lines.append(MintMetric(metric_name, col.value, dimensions))
                    else:
                        mint_lines.append(MintMetric(metric_name, col.value))

                    break  # We ignore the rest of the rows in this case
            except Exception as e:
                self.logger.exception(f"{query.name} - Could not parse rows: {e}")

        return mint_lines

    def read_results_from_file(self) -> Result:
        """
        Reads the <endpoint_id>.json file inside the results folder
        and parses the data into a models.Result object
        """
        results_file = Path(self.temp_folder, "results", f"{self.activation.entity_id}.json")
        if results_file.exists():
            with open(results_file, "r") as f:
                try:
                    return Result.from_json(json.load(f))
                except Exception as e:
                    self.logger.error(f"Could not parse json from {results_file}: {e}")
                    raise NothingToReportException(f"Error: {e}")
        return Result()

    def validate_parameters(self):

        # Check that at least one query was configured
        query_names = [self.config.get(f"query_{i}_name", "") for i in range(1, 11)]
        # Count the numboer of query names that are not empty
        num_queries = len([q for q in query_names if q])
        self.logger.info(f"{num_queries} queries configured")
        if num_queries == 0:
            raise ConfigException("At least one query must be configured")


def chunks(elements, size):
    """
    Breaks a list into smaller lists of size elements
    """
    for i in range(0, len(elements), size):
        yield elements[i : i + size]


def slugify(value: str) -> str:
    """
    Converts a string into an url safe string
    """
    value = unicodedata.normalize("NFKC", value)
    value = re.sub(r"[:.-/\s]+", "_", value)
    value = re.sub(r"[^\w\s-]", "", value)
    return value.lower()


def create_or_get_temp_folder():
    temp_folder = Path(tempfile.gettempdir(), "dynatrace-ag-dbquery")
    if not temp_folder.exists():
        os.mkdir(temp_folder)

    other_paths = [
        Path(temp_folder, "config"),
        Path(temp_folder, "log"),
        Path(temp_folder, "results"),
    ]

    for path in other_paths:
        if not path.exists():
            os.mkdir(path)

    return temp_folder
